#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/i82489.h>
#include <asm/fixmap.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/atomic.h>
#include <asm/desc.h>

#include <asm/system.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/atomic.h>
#include <asm/desc.h>
#include <asm/io.h>

#include <rtai.h>
#include <rtai_sched.h>

static int strsiz(char *str)
{
	int i = 0;
	while (str[i++]);
	return i - 1;
}
static char *strs[] = { "Joey ", "Johnny ", "Dee Dee ", "Marky " };
static char sync_str[] = "gabba\n";

#define NUM_TASKS (sizeof(strs) / sizeof(char *))

RT_TASK start_task, tasks[NUM_TASKS];

SEM sems[NUM_TASKS], sync_sem, prio_sem;

MBX mbx_in, mbx_out;

#define DBG_PRINT_SETUP         SEM print_sem;
#define DBG_PRINT_INIT          rt_sem_init(&print_sem, 1);
#define TAKE_PRINT              rt_sem_wait(&print_sem);
#define GIVE_PRINT              rt_sem_signal(&print_sem);

DBG_PRINT_SETUP

/*
 *  Each task waits to receive the semaphore, prints its string, and
 *  passes the semaphore to the next task.  Then it sends a sync semaphore,
 *  and waits for another semaphore, and this time displays it in
 *  priority order.  Finally, message queues are tested.
 */

static void task_code(int task_no)
{
	int i, ret;
	char buf[20];
	for (i = 0; i < 5; ++i)	{
		rt_sem_wait(&sems[task_no]);
		printk(strs[task_no]);
		if (task_no == NUM_TASKS-1)
		printk("\n");
		rt_sem_signal(&sems[(task_no + 1) % NUM_TASKS]);
	}
	rt_sem_signal(&sync_sem);
	rt_sem_wait(&prio_sem);
	printk(strs[task_no]);
	rt_sleep(nano2count(1E9));
	rt_sem_wait_timed(&prio_sem, nano2count((task_no + 1)*1E9));
	printk(strs[task_no]);
	rt_sem_signal(&sync_sem);

/* message queue stuff */
	if ((ret = rt_mbx_receive(&mbx_in, buf, strsiz(strs[task_no]))) != 0) {
		printk("rt_mq_receive() failed with %d\n", ret);
	}
	TAKE_PRINT; 
	printk("\nreceived by task %d ", task_no);
	buf[strsiz(strs[task_no])] = 0; 
	printk(buf); 
	GIVE_PRINT;
	rt_mbx_send(&mbx_out, strs[task_no], strsiz(strs[task_no]));
/* test receive timeout */
	rt_sem_wait(&sync_sem);
	if (rt_mbx_receive(&mbx_in, buf, nano2count(task_no + 1))) {
		TAKE_PRINT;
		printk("task %d timed out.\n", task_no); 
		GIVE_PRINT;
	}

	rt_task_suspend(&tasks[task_no]);
}

/*
 * initialization task
 */
static void start_task_code(int notused)
{
	int i;
	DBG_PRINT_INIT;
	if (rt_mbx_init(&mbx_in, NUM_TASKS*20) || rt_mbx_init(&mbx_out, NUM_TASKS*20)) {
		printk("could not create message queue\n");
		return;
	}
	for (i = 0; i < NUM_TASKS; ++i) {
		rt_sem_init(&sems[i], 0);
		if (rt_task_init(&tasks[i], task_code, i, 3000, NUM_TASKS-i, 0, 0) != 0) {
			printk("rt_task_ipc_init failed\n");
			return;
		}
		rt_task_resume(&tasks[i]);
	}     
  /* create the sync semaphore */
	rt_sem_init(&sync_sem, 0);
  /* create the priority-test semaphore */
	rt_sem_init(&prio_sem, 0);
  /* pass the semaphore to the first task */
	rt_sem_signal(&sems[0]);
  /* wait for each task to send the sync semaphore */
	for (i = 0; i < NUM_TASKS; ++i) {
		rt_sem_wait(&sync_sem);
	}
	printk(sync_str);
  /* post the priority-test semaphore -- the tasks should then run */
  /* in priority order */
	for (i = 0; i < NUM_TASKS; ++i) {
		rt_sem_signal(&prio_sem);
	}
	printk("\n");
	for (i = 0; i < NUM_TASKS; ++i) {
		rt_sem_wait(&sync_sem);
	}
	printk(sync_str);

  /* now, test message queues */
	printk("testing message queues\n");
	for (i = 0; i < NUM_TASKS; ++i) {
		if (rt_mbx_send(&mbx_in, strs[i], strsiz(strs[i]))) {
			printk("rt_mbx_send() failed\n");
		}
	}
	for (i = 0; i < NUM_TASKS; ++i) {
		char buf[20];
		rt_mbx_receive(&mbx_out, buf, strsiz(strs[i]));
		buf[strsiz(strs[i])] = 0;
		TAKE_PRINT; 
		printk("\nreceived from mbx_out: %s", buf); 
		GIVE_PRINT;
	}
	for (i = 0; i < NUM_TASKS; ++i) {
		rt_sem_signal(&sync_sem);
	}
	TAKE_PRINT; 
	printk("\ninit task complete\n"); 
	GIVE_PRINT;

  /* nothing more for this task to do */
	rt_task_suspend(&start_task);
}

int init_module(void)
{
	rt_set_oneshot_mode();
	start_rt_timer(nano2count(1E7));
	if (rt_task_init(&start_task, start_task_code, 0, 3000, 10, 0, 0) != 0) {
		printk("Could not start init task\n");
	}
	rt_task_resume(&start_task);
	return 0;
}

void cleanup_module(void)
{
	int i;
	stop_rt_timer();
	rt_sleep(1E8);
	for (i = 0; i < NUM_TASKS; ++i) {
		rt_sem_delete(&sems[i]);
		rt_task_delete(&tasks[i]);
	}
	rt_sem_delete(&sync_sem);
	rt_mbx_delete(&mbx_in);
	rt_mbx_delete(&mbx_out);
	rt_task_delete(&start_task);
}
