/*
COPYRIGHT (C) 1999  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/


#define LXRT_MODULE
#include <linux/version.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/wrapper.h>

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_lxrt.h>


#define __STR(x) #x
#define STR(x) __STR(x)

//#define RET_FROM_INTR  // To force all the system processing stuff.
#ifdef RET_FROM_INTR

static void rtai_lxrt_handler(void)
{
	__asm__ __volatile__ (" \
	pushl %eax;\n\t \
	cld; pushl %es; pushl %ds; pushl %eax; pushl %ebp;\n\t \
	pushl %edi; pushl %esi; pushl %edx; pushl %ecx;\n\t \
	pushl %ebx; pushl %edx; pushl %eax;\n\t \
	movl $" STR(__KERNEL_DS) ",%ebx; movl %bx,%ds; movl %bx,%es");

	__asm__ __volatile__ ("call "SYMBOL_NAME_STR(lxrt_handler));

	__asm__ __volatile__ (" \
	addl $8,%esp; movl %edx,8(%esp); movl %eax,24(%esp);\n\t \
	jmp ret_from_intr");
}

#else

static void rtai_lxrt_handler(void)
{
	__asm__ __volatile__ (" \
	cld; pushl %es; pushl %ds; pushl %ebp;\n\t \
	pushl %edi; pushl %esi; pushl %ecx;\n\t \
	pushl %ebx; pushl %edx; pushl %eax;\n\t \
	movl $" STR(__KERNEL_DS) ",%ebx; movl %bx,%ds; movl %bx,%es");

	__asm__ __volatile__ ("call "SYMBOL_NAME_STR(lxrt_handler));

	__asm__ __volatile__ (" \
	addl $8,%esp; popl %ebx; popl %ecx; popl %esi;\n\t \
	popl %edi; popl %ebp; popl %ds; popl %es; iret");
}

#endif

/*     The macro below must be >= the max number of all objects to be 
    registered and shared (objects are: rt_tasks, semaphores, mailboxes).  */
#define MAX_SLOTS  100

static struct { unsigned long name; void *adr; } lxrt_list[MAX_SLOTS + 1] = { {0, 0}, };
static spinlock_t list_lock = SPIN_LOCK_UNLOCKED;

static inline int registr(unsigned long name, void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (!lxrt_list[slot].name) {
			lxrt_list[slot].name = name;
			lxrt_list[slot].adr  = adr;
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline int drg_on_name(unsigned long name)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].name == name) {
			lxrt_list[slot] = lxrt_list[0];
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline int drg_on_adr(void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].adr == adr) {
			lxrt_list[slot] = lxrt_list[0];
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline unsigned long get_name(void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].adr == adr) {
		rt_spin_unlock_irqrestore(flags, &list_lock);
		return lxrt_list[slot].name;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline void *get_adr(unsigned long name)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].name == name) {
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return lxrt_list[slot].adr;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

int rt_register(unsigned long name, void *adr)
{
	return registr(name, adr);
}

int rt_drg_on_name(unsigned long name)
{
	return drg_on_name(name);
} 

int rt_drg_on_adr(void *adr)
{
	return drg_on_adr(adr);
} 

unsigned long rt_get_name(void *adr)
{
	return get_name(adr);
} 

void *rt_get_adr(unsigned long name)
{
	return get_adr(name);
}

#define LXRT_RESUME(x)  {extern void emuser_trxl(RT_TASK *); emuser_trxl((x));}
#define LXRT_SUSPEND(x) {extern void (*dnepsus_trxl)(void); (*dnepsus_trxl)();}

// BE CAREFULL TO KEEP THE 3 MACROS BELOW UP TO DATE WITH THE SCHEDULERS!
// WATCH OUT for:    1) Largest stack size required by any scheduler function;
//		     2) Max size of messages expected in mail boxes;
//		     3) Max number of 4 bytes args in scheduler functions.
// Note that the first two are just defaults. You are in full control of the 
// stack size at task init and the mail box size is automatically updated when
// needed, with a GFP_KERNEL non atomic kmalloc. 

#define STACK_SIZE    512	// Default stack size.
#define MSG_SIZE      256	// Default max message size.
#define MAX_ARGS       10	// Max 4 bytes args in rtai functions.

/*       !!! ATTENTION: KEEP THE MACRO BELOW A POWER OF 2, ALWAYS !!! 
         Must be >= the max number of expected real time Linux tasks.       */
#define MAX_SRQ  128

static struct {int srq, in, out; struct wait_queue **waitq[MAX_SRQ]; spinlock_t lock;} sysrq;

typedef union { unsigned long args[MAX_ARGS]; } ARGS;

static void lxrt_suspend(ARGS args)
{
#define rt_task ((RT_TASK *)(args.args[MAX_ARGS - 1]))
	__asm__ __volatile__ ("" : "=A" (rt_task->retval) );
	rt_global_cli();
	if ((rt_task->lnxtsk.task)->state == TASK_INTERRUPTIBLE) {
		sysrq.waitq[sysrq.in] = &(rt_task->waitq);
		sysrq.in = (sysrq.in + 1) & (MAX_SLOTS - 1);
		rt_pend_linux_srq(sysrq.srq);
	}
	rt_task->state = 0;
	LXRT_SUSPEND(rt_task);
	printk("!!! GREAT PAIN, THIS POINT CANNOT BE REACHED !!!\n"); 
#undef rt_task
}

static void lxrt_global_sti(void)
{
	if (test_and_clear_bit(hard_cpu_id(), locked_cpus)) {
		clear_bit(31, locked_cpus);
	}
	hard_sti();
}

static inline long long lxrt_resume(void *fun, int narg, int *arg, int type, RT_TASK *rt_task)
{
	int msg_size, *msg_adr;

	copy_from_user(rt_task->stack_top, arg, (narg - 1)*sizeof(int));

/* Here type > 0 means any messaging with the need of copying from/to user 
   space. My knowledge of Linux memory menagment has led to this mess. 
   Whoever can do it better is warmly welcomed.                             */
	msg_adr = 0;
	msg_size = 0;
	if (type > 0) {
		if (type > 2) {
// Mail box message.
			if ((msg_size = *(rt_task->stack_top + 2)) > rt_task->max_msg_size) {
				kfree(rt_task->msg_buf);
				rt_task->max_msg_size = ((*(rt_task->stack_top + 2))*120 + 50)/100;
				rt_task->msg_buf = kmalloc(rt_task->max_msg_size, GFP_KERNEL);
			}
			msg_adr = (int *)(*(rt_task->stack_top + 1));
			*(rt_task->stack_top + 1) = (int)rt_task->msg_buf;
			if (type == 3) {
				copy_from_user(rt_task->msg_buf, msg_adr, msg_size);
			}
		} else {
// Inter tasks message.
			msg_size = sizeof(int);
			msg_adr = (int *)(*(rt_task->stack_top + type));
			*(rt_task->stack_top + type) = (int)rt_task->msg_buf;
		}
	}
/* End of messaging mess */
	*(rt_task->stack_top - 1) = (int)lxrt_suspend;
	*(rt_task->stack_top - 2) = (int)fun;
	*(rt_task->stack = rt_task->stack_top - 3) = (int)lxrt_global_sti;
	rt_global_cli();
	LXRT_RESUME(rt_task);
	rt_global_cli();
	if (rt_task->state) {
		current->state = TASK_INTERRUPTIBLE;
		rt_global_sti();
		schedule();
	} else {
		rt_global_sti();
	}
/* A trashing of the comment about messaging mess at the beginning */
	if (type > 0 && type != 3) {
		copy_to_user(msg_adr, rt_task->msg_buf, msg_size);
	}
	return rt_task->retval;
}

long long lxrt_handler(unsigned int lxsrq, void *arg)
{
	union {unsigned long name; RT_TASK *rt_task; SEM *sem; MBX *mbx; } arg0;
	unsigned long name;
	int srq;

	get_user(arg0.name, (unsigned long *)arg);
	if ((srq = SRQ(lxsrq)) < GET_ADR) {
		if (rt_fun[srq].type) {
			return lxrt_resume(rt_fun[srq].fun, NARG(lxsrq), &((int *)arg)[1], rt_fun[srq].type - 1, arg0.rt_task);
		}

		switch (NARG(lxsrq)) {
			case 0:
				return ((long long (*)(void))rt_fun[srq].fun)();
			case 1:
				return ((long long (*)(int))rt_fun[srq].fun)(arg0.name);
			case 2: {
				int arg1;
				get_user(arg1, &((int *)arg)[1]);
				return ((long long (*)(int, int))rt_fun[srq].fun)(arg0.name, arg1);
			}
		}
	}

	switch (srq) {

		case GET_ADR: {
			return (unsigned long)get_adr(arg0.name);
		}

		case GET_NAME: {
			return get_name((void *)arg0.name);
		}

		case TASK_INIT: {
			int prio, stack_size, max_msg_size;
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			get_user(prio, &((int *)arg)[1]);
			if (prio > RT_LOWEST_PRIORITY) {
				prio = RT_LOWEST_PRIORITY;
			}
			get_user(stack_size, &((int *)arg)[2]);
			if (!stack_size) {
				stack_size = STACK_SIZE;
			}
			get_user(max_msg_size, &((int *)arg)[3]);
			if (!max_msg_size) {
				max_msg_size = MSG_SIZE;
			}
			if (!(arg0.rt_task = kmalloc(sizeof(RT_TASK), GFP_KERNEL)) ||
			    !(arg0.rt_task->msg_buf = kmalloc(max_msg_size, GFP_KERNEL))) {
				return 0;
			}
			if (!rt_task_init(arg0.rt_task, 0, 0, stack_size, prio, 0, 0)) {
				*(arg0.rt_task->stack) = (int)arg0.rt_task;
				arg0.rt_task->stack_top = arg0.rt_task->stack - MAX_ARGS;
				arg0.rt_task->max_msg_size = max_msg_size;
				arg0.rt_task->state = 0;
				arg0.rt_task->waitq = 0;
				arg0.rt_task->lnxtsk.task = current;
				__add_wait_queue(&(arg0.rt_task->waitq), &(arg0.rt_task->lnxtsk));
				if (registr(name, arg0.rt_task)) {
					return arg0.name;
				} else {
					kfree(arg0.rt_task);
				}
			}
			return 0;
		}

		case TASK_DELETE: {
			if (rt_task_delete(arg0.rt_task)) {
				return -EFAULT;
			}
			kfree(arg0.rt_task->msg_buf);
			kfree(arg0.rt_task);
			return drg_on_adr(arg0.rt_task);
		}

		case SEM_INIT: {
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			if ((arg0.sem = kmalloc(sizeof(SEM), GFP_KERNEL))) {
				int ini_cnt;
				get_user(ini_cnt, &((int *)arg)[1]);
				rt_sem_init(arg0.sem, ini_cnt);
				if (registr(name, arg0.sem)) {
					return arg0.name;
				} else {
					kfree(arg0.sem);
				}
			}
			return 0;
		}

		case SEM_DELETE: {
			if (rt_sem_delete(arg0.sem)) {
				return -EFAULT;
			}
			kfree(arg0.sem);
			return drg_on_adr(arg0.sem);
		}

		case MBX_INIT: {
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			if ((arg0.mbx = kmalloc(sizeof(MBX), GFP_KERNEL))) {
				int size;
				get_user(size, &((int *)arg)[1]);
				rt_mbx_init(arg0.mbx, size);
				if (registr(name, arg0.mbx)) {
					return arg0.name;
				} else {
					kfree(arg0.mbx);
				}
			}
			return 0;
		}

		case MBX_DELETE: {
			if (rt_mbx_delete(arg0.mbx)) {
				return -EFAULT;
			}
			kfree(arg0.mbx);
			return drg_on_adr(arg0.mbx);
		}
	}
	return 0;
}

static void lxrt_srq_handler(void)
{
	while (sysrq.out != sysrq.in) {
		wake_up_interruptible(sysrq.waitq[sysrq.out]);
		sysrq.out = (sysrq.out + 1) & (MAX_SLOTS - 1);
	}
}

static struct desc_struct sidt;

int init_module(void)
{
	sidt = rt_set_full_intr_vect(RTAI_LXRT_VECTOR, 15, 3, rtai_lxrt_handler);
	if ((sysrq.srq = rt_request_srq(nam2num("LXRT"), lxrt_srq_handler, 0)) < 0) {
		printk("LXRT: no sysrq available.\n");
		return sysrq.srq;
	}
	sysrq.in = sysrq.out = sysrq.lock.lock = 0;
	return 0 ;
}

void cleanup_module(void)
{
	if (rt_free_srq(sysrq.srq) < 0) {
		printk("LXRT: sysrq %d illegal or already free.\n", sysrq.srq);
	}
	rt_reset_full_intr_vect(RTAI_LXRT_VECTOR, sidt);
	return;
}
